/****************************************************************************
 Include Section
*****************************************************************************/

#include "global.h"
#include "init_platform.h"

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __clear_reg;

.GLOBAL __full_on;
.GLOBAL __full_on.dyn;

.GLOBAL __get_vco_hz;
.GLOBAL __get_cclk_hz;
.GLOBAL __get_sclk_hz;

#if defined SDRAM_SIZE
    .GLOBAL __sdram_en;
    .GLOBAL __sdram.self_refresh.enter;
    .GLOBAL __sdram.self_refresh.leave;
#if !defined __ADSPBF54x__
    .GLOBAL __get_rdiv;
#endif
#endif

.GLOBAL __sysreset;
.GLOBAL __softreset;

.GLOBAL __trace_unit.enable;
.GLOBAL __trace_unit.disable;

.GLOBAL __profiling.enable;
.GLOBAL __profiling.disable;

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION program;


/****************************************************************************
 Function:              __clear_reg
 Description:           set all core register to zero
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:0,P5:0
                        A1:0,ASTAT
                        B3:0,I3:0,M3:0,L3:0
                        LC1:0,LT1:0,LB1:0
 Global Registers Used: none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__clear_reg:

    link 0;

    /* Clear General Registers (genreg): [R7:0], [P5:0], SP, FP, [A1:0] */
    R0 = 0; R1 = 0; R2 = 0; R3 = 0; R4 = 0; R5 = 0; R6 = 0; R7 = 0;
    P0 = 0; P1 = 0; P2 = 0; P3 = 0; P4 = 0; P5 = 0;
    A0 = 0; A1 = 0; ASTAT = R7;
    /* Clear DAG Registers (dagreg): [B3:0], [I3:0], [M3:0], [L3:0] */
    B0 = 0; B1 = 0; B2 = 0; B3 = 0;
    I0 = 0; I1 = 0; I2 = 0; I3 = 0;
    M0 = 0; M1 = 0; M2 = 0; M3 = 0;
    L0 = 0; L1 = 0; L2 = 0; L3 = 0;
    /* FYI: System Registers (sysreg): ASTAT, SEQSTAT, SYSCFG, RETI, RETX, RETN, RETE, RETS,
    LC0 and LC1, LT0 and LT1, LB0 and LB1, CYCLES, CYCLES2, and EMUDAT */
    LC0 = P5; LT0 = P5; LB0 = P5;
    LC1 = P5; LT1 = P5; LB1 = P5;
    /* Cycle Counter */
    CYCLES = R7;
    CYCLES2 = R7;

    unlink;

    rts;
__clear_reg.end:


DPM:


#if defined __USEBFSYSCONTROL__


/****************************************************************************
 Name:                  __full_on
 Description:           bring processor to full on mode
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:R0,P5:P0
 Global Registers Used: none
 HINT:                  As a C-Callable function is called, take care of the
                        C/C++ run-time environment scratch registers
*****************************************************************************/

__full_on:

    link sizeof(ADI_SYSCTRL_VALUES)+2;
    [--SP] = (R7:0,P5:0);
    SP += -12;


    R0.L = PLL_CTL_VAL;
    w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllCtl)] = R0;
    R0.L = PLL_LOCKCNT_VAL;
    w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllLockCnt)] = R0;

    #if defined (__ADSPBF51x__) || defined (__ADSPBF526_FAMILY__)
        /* no internal voltage regulator */
    #else
        R0.L = VR_CTL_VAL;
        w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwVrCtl)] = R0;
    #endif

    #if defined __WORKAROUND_05000440__

        R0.L = PLL_DIV_VAL;
        IMM32(P5,PLL_DIV);
        w[P5] = R0;

        #if defined (__ADSPBF51x__) || defined (__ADSPBF526_FAMILY__)
            R0 = (                 SYSCTRL_EXTVOLTAGE | SYSCTRL_PLLCTL |                  SYSCTRL_LOCKCNT | SYSCTRL_WRITE );
        #else
            R0 = ( SYSCTRL_VRCTL | SYSCTRL_INTVOLTAGE | SYSCTRL_PLLCTL |                  SYSCTRL_LOCKCNT | SYSCTRL_WRITE );
        #endif

    #else

        R0.L = PLL_DIV_VAL;
        w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllDiv)] = R0;

        #if defined (__ADSPBF51x__) || defined (__ADSPBF526_FAMILY__)
            R0 = (                 SYSCTRL_EXTVOLTAGE | SYSCTRL_PLLCTL | SYSCTRL_PLLDIV | SYSCTRL_LOCKCNT | SYSCTRL_WRITE );
        #else
            R0 = ( SYSCTRL_VRCTL | SYSCTRL_INTVOLTAGE | SYSCTRL_PLLCTL | SYSCTRL_PLLDIV | SYSCTRL_LOCKCNT | SYSCTRL_WRITE );
        #endif

    #endif

    R1 = FP;
    R1 += -sizeof(ADI_SYSCTRL_VALUES);
    R2 = 0 (z);
    IMM32(P4,BFROM_SYSCONTROL);
    call(P4);


    SP += 12;
    (R7:0,P5:0) = [SP++];
    unlink;
    rts;

__full_on.end:


/****************************************************************************
 Name:                  __full_on.dyn
 Description:           bring processor to full on mode
 Input Parameters:      R0 = VR_CTL
                        R1 = PLL_DIV
                        R2 = PLL_CTL
 Return Parameters:     none
 Registers Used:        R7:6,P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__full_on.dyn:

    link sizeof(ADI_SYSCTRL_VALUES)+2;
    [--SP] = (R7:0,P5:0);
    SP += -12; /* Bubble: Always allocate at least 12 bytes on the stack for outgoing arguments, even if the function being called requires less than this. */


    #if defined SDRAM_SIZE
        call __sdram.self_refresh.enter;
    #endif


    #if defined (__ADSPBF51x__) || defined (__ADSPBF52EVEN__)
        /* no internal voltage regulator */
    #else
        w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwVrCtl)] = R0;
    #endif
    w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllCtl)] = R2;
    w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllDiv)] = R1;
    R0.L = PLL_LOCKCNT_VAL;
    w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllLockCnt)] = R0;

    #if defined (__ADSPBF51x__) || defined (__ADSPBF52EVEN__)
        R0 = (                 SYSCTRL_EXTVOLTAGE | SYSCTRL_PLLCTL | SYSCTRL_PLLDIV | SYSCTRL_LOCKCNT | SYSCTRL_WRITE );
    #else
        R0 = ( SYSCTRL_VRCTL | SYSCTRL_INTVOLTAGE | SYSCTRL_PLLCTL | SYSCTRL_PLLDIV | SYSCTRL_LOCKCNT | SYSCTRL_WRITE );
    #endif
    R1 = FP;
    R1 += -sizeof(ADI_SYSCTRL_VALUES);
    R2 = 0 (z);
    IMM32(P4,BFROM_SYSCONTROL);
    call(P4);


    #if defined SDRAM_SIZE
        #if !defined __ADSPBF54x__
            call __get_rdiv;
            w[P5 + lo(EBIU_SDRRC)] = R0;
        #endif
        call __sdram.self_refresh.leave;
    #endif


    SP += 12;
    (R7:0,P5:0) = [SP++];
    unlink;
    rts;

__full_on.dyn.end:


/****************************************************************************
 Name:                  __get_vco_hz
 Description:           get PLL VCO frequency in Hz
 Input Parameters:      none
 Return Parameters:     R0 = VCO
 Registers Used:        R7:R1,P5:P0
 Global Registers Used: none
 HINT:                  As a C-Callable function is called, take care of the
                        C/C++ run-time environment scratch registers
*****************************************************************************/

__get_vco_hz:

    link sizeof(ADI_SYSCTRL_VALUES)+2; /* 10 = sizeof(ADI_SYSCTRL_VALUES), but factor 4 required */
    [--SP] = (R7:1,P5:0);
    SP += -12; /* bubble: Always allocate at least 12 bytes on the stack for outgoing arguments, even if the function being called requires less than this. */
    IMM32(P5,SYS_MMR_BASE);


    /* get current MSEL */
    R0 = ( SYSCTRL_INTVOLTAGE | SYSCTRL_PLLCTL | SYSCTRL_READ );
    R1 = FP;
    R1 += -sizeof(ADI_SYSCTRL_VALUES);
    R2 = 0;
    IMM32(P4,BFROM_SYSCONTROL);
    call (P4);
    R7 = w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllCtl)]; /* R7 = w[P5 + lo(PLL_CTL)] (z); */
    R0 = MSEL (z);
    R0 = R0 & R7;
    R0 >>= 9;               /* R0 = current MSEL */

    /* get current VCO */
    IMM32(R6,CLKIN_Hz);     /* R6 = CLKIN [Hz] */
    R0 *= R6;               /* R0 = MSEL * CLKIN */
    R6 = DF (z);            /* R6 = DF bit */
    R6 = R6 & R7;           /* R6 = DF bit [1/0] */
    R0 >>= R6;              /* R0 = current VCO [Hz] */


    SP += 12; /* unbubble */
    (R7:1,P5:0) = [SP++];
    unlink;
    rts;

__get_vco_hz.end:


/****************************************************************************
 Name:                  __get_cclk_hz
 Description:           get core clock frequency in Hz
 Input Parameters:      none
 Return Parameters:     R0 = CCLK
 Registers Used:        R7:R1,P5:P0,ASTAT
 Global Registers Used: none
 HINT:                  As a C-Callable function is called, take care of the
                        C/C++ run-time environment scratch registers
*****************************************************************************/

__get_cclk_hz:

    link sizeof(ADI_SYSCTRL_VALUES)+2; /* 10 = sizeof(ADI_SYSCTRL_VALUES), but factor 4 required */
    [--SP] = ASTAT;
    [--SP] = (R7:1,P5:0);
    SP += -12; /* bubble: Always allocate at least 12 bytes on the stack for outgoing arguments, even if the function being called requires less than this. */
    IMM32(P5,SYS_MMR_BASE);


    R0 = ( SYSCTRL_INTVOLTAGE | SYSCTRL_PLLCTL | SYSCTRL_READ );
    R1 = FP;
    R1 += -sizeof(ADI_SYSCTRL_VALUES);
    R2 = 0;
    IMM32(P4,BFROM_SYSCONTROL);
    call (P4);
    R7 = w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllCtl)]; /* R7 = w[P5 + lo(PLL_CTL)] (z); */
    CC = bittst(R7,bitpos(BYPASS));
    IMM32(R0,CLKIN_Hz);
    if CC jump get_cclk_hz.skip;

    call __get_vco_hz;        /* R0 = current VCO [Hz] */

    /* get current CSEL */
    R7 = w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllDiv)]; /* R7 = w[P5 + lo(PLL_DIV)] (z); */
    R6 = CSEL (z);
    R7 = R7 & R6;
    R7 >>= 4;               /* R7 = current CSEL */

    R0 >>= R7;              /* R0 = current CCLK */


    get_cclk_hz.skip:
    SP += 12; /* unbubble */
    (R7:1,P5:0) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__get_cclk_hz.end:


/****************************************************************************
 Name:                  __get_sclk_hz
 Description:           get system clock frequency in Hz
 Input Parameters:      none
 Return Parameters:     R0 = SCLK
 Registers Used:        R7:R1,P5:P0,ASTAT
 Global Registers Used: none
 HINT:                  As a C-Callable function is called, take care of the
                        C/C++ run-time environment scratch registers
*****************************************************************************/

__get_sclk_hz:

    link sizeof(ADI_SYSCTRL_VALUES)+2; /* 10 = sizeof(ADI_SYSCTRL_VALUES), but factor 4 required */
    [--SP] = ASTAT;
    [--SP] = (R7:1,P5:0);
    SP += -12; /* bubble: Always allocate at least 12 bytes on the stack for outgoing arguments, even if the function being called requires less than this. */
    IMM32(P5,SYS_MMR_BASE);


    R0 = ( SYSCTRL_INTVOLTAGE | SYSCTRL_PLLCTL | SYSCTRL_PLLDIV | SYSCTRL_READ );
    R1 = FP;
    R1 += -sizeof(ADI_SYSCTRL_VALUES);
    R2 = 0;
    IMM32(P4,BFROM_SYSCONTROL);
    call (P4);
    R7 = w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllCtl)]; /* R7 = w[P5 + lo(PLL_CTL)] (z); */
    CC = bittst(R7,bitpos(BYPASS));
    IMM32(R0,CLKIN_Hz);
    if CC jump get_sclk_hz.skip;

    call __get_vco_hz;        /* R0 = current VCO [Hz] */

    /* get current SSEL */
    R1 = w[FP+-sizeof(ADI_SYSCTRL_VALUES)+offsetof(ADI_SYSCTRL_VALUES,uwPllDiv)]; /* R1 = w[P5 + lo(PLL_DIV)] (z); */
    R7 = SSEL (z);
    R1 = R1 & R7;           /* R1 = current SSEL */

    /* get current SCLK */
    udiv32(); /* Registers passed: Dividend = R0, Divisor = R1, Quotient -> R0 */
                            /* R0 = current SCLK */


    get_sclk_hz.skip:
    SP += 12; /* unbubble */
    (R7:1,P5:0) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__get_sclk_hz.end:


#else /* !defined __USEBFSYSCONTROL__ */


/****************************************************************************
 Name:                  __full_on
 Description:           bring processor to full on mode
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__full_on:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /*****************************************************************************************
     Description:   Dynamic Power Managment

     Important Notice:
     Depending on increasing/decreasing speed (core clock (CCLK) and/or system clock (SCLK)),
     the order of the registers PLL_CTL, PLL_DIV (and VR_CTL) have to be adapted!
     - Increase speed: 1.  VR_CTL, 2. PLL_DIV, 3. PLL_CTL
     - Decrease speed: 1. PLL_CTL, 2. PLL_DIV, 3.  VR_CTL
     If SDRAM is in use, the value for RDIV in the EBIU_SDRRC register has to be re-calculated
    *****************************************************************************************/


    /**************************
    * PLL Lock Count Register *
    **************************/

    R7.L = PLL_LOCKCNT_VAL;
    w[P5 + lo(PLL_LOCKCNT)] = R7;


    /*****************************************************************************
     Description:   Configure Voltage Regulator Control Register

     IMPORTANT
       NOTICE:      The same value should not be written twice
                    to the VR_CTL register
    ******************************************************************************/

    R6 = w[P5 + lo(VR_CTL)] (z);
    R7 = VR_CTL_VAL (z);
    CC = R6 == R7;
    if CC jump skip_vr_ctl;

    cli R6;                     /* disable interrupts */
    w[P5 + lo(VR_CTL)] = R7;    /* Apply VR_CTL changes */
    idle;                       /* drain pipeline and send core into IDLE. Required for changing voltage settings */
    sti R6;                     /* re-enable interrupts after wakeup */

    skip_vr_ctl: nop;


    /*********************************
    * PLL Divide Register            *
    * Can be configured on the fly   *
    * First configure divider,       *
    * than configure MSEL in PLL_CTL *
    *********************************/

    R7.L = PLL_DIV_VAL;
    w[P5 + lo(PLL_DIV)] = R7;


    /*****************************************************************************
     Description:   Configure PLL Control Register
     IMPORTANT
       NOTICE:      The same value should not be written twice
                    to the PLL_CTL register
     IMPORTANT
        NOTICE
     for BF561:     Core B must be in idle state during this programming sequence
    ******************************************************************************/

    R6 = w[P5 + lo(PLL_CTL)] (z);
    R7 = PLL_CTL_VAL (z);
    CC = R6 == R7;
    if CC jump skip_pll_ctl;

    cli R6;                     /* disable interrupts */
    w[P5 + lo(PLL_CTL)] = R7;   /* Apply PLL_CTL changes */
    idle;                       /* drain pipeline and send core into IDLE. Required for changing voltage settings */
    sti R6;                     /* re-enable interrupts after wakeup */

    skip_pll_ctl: nop;


    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__full_on.end:


/****************************************************************************
 Name:                  __full_on.dyn
 Description:           bring processor to full on mode
 Input Parameters:      R0 = VR_CTL
                        R1 = PLL_DIV
                        R2 = PLL_CTL
 Return Parameters:     none
 Registers Used:        R7:6,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__full_on.dyn:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    #if defined SDRAM_SIZE
        call __sdram.self_refresh.enter;
    #endif

    /*****************************************************************************************
     Description:   Dynamic Power Managment

     Important Notice:
     Depending on increasing/decreasing speed (core clock (CCLK) and/or system clock (SCLK)),
     the order of the registers PLL_CTL, PLL_DIV (and VR_CTL) have to be adapted!
     - Increase speed: 1.  VR_CTL, 2. PLL_DIV, 3. PLL_CTL
     - Decrease speed: 1. PLL_CTL, 2. PLL_DIV, 3.  VR_CTL
     If SDRAM is in use, the value for RDIV in the EBIU_SDRRC register has to be re-calculated
    *****************************************************************************************/


    /**************************
    * PLL Lock Count Register *
    **************************/

    R7.L = PLL_LOCKCNT_VAL;
    w[P5 + lo(PLL_LOCKCNT)] = R7;


    /*****************************************************************************
     Description:   Configure Voltage Regulator Control Register

     IMPORTANT
       NOTICE:      The same value should not be written twice
                    to the VR_CTL register
    ******************************************************************************/

    R6 = w[P5 + lo(VR_CTL)] (z);
    CC = R6 == R0;
    if CC jump dyn.skip_vr_ctl;

    cli R6;                     /* disable interrupts */
    w[P5 + lo(VR_CTL)] = R0;    /* Apply VR_CTL changes */
    idle;                       /* drain pipeline and send core into IDLE. Required for changing voltage settings */
    sti R6;                     /* re-enable interrupts after wakeup */

    dyn.skip_vr_ctl: nop;


    /*********************************
    * PLL Divide Register            *
    * Can be configured on the fly   *
    * First configure divider,       *
    * than configure MSEL in PLL_CTL *
    *********************************/

    w[P5 + lo(PLL_DIV)] = R1;


    /*****************************************************************************
     Description:   Configure PLL Control Register
     IMPORTANT
       NOTICE:      The same value should not be written twice
                    to the PLL_CTL register
     IMPORTANT
        NOTICE
     for BF561:     Core B must be in idle state during this programming sequence
    ******************************************************************************/

    R6 = w[P5 + lo(PLL_CTL)] (z);
    CC = R6 == R2;
    if CC jump dyn.skip_pll_ctl;

    cli R6;                     /* disable interrupts */
    w[P5 + lo(PLL_CTL)] = R2;   /* Apply PLL_CTL changes */
    idle;                       /* drain pipeline and send core into IDLE. Required for changing voltage settings */
    sti R6;                     /* re-enable interrupts after wakeup */

    dyn.skip_pll_ctl: nop;


    #if defined SDRAM_SIZE
        #if !defined __ADSPBF54x__
            call __get_rdiv;
            w[P5 + lo(EBIU_SDRRC)] = R0;
        #endif
        call __sdram.self_refresh.leave;
    #endif


    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__full_on.dyn.end:


/****************************************************************************
 Name:                  __get_vco_hz
 Description:           get PLL VCO frequency in Hz
 Input Parameters:      none
 Return Parameters:     R0 = VCO
 Registers Used:        R7:6,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__get_vco_hz:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /* get current MSEL */
    R7 = w[P5 + lo(PLL_CTL)] (z);
    R0 = MSEL (z);
    R0 = R0 & R7;
    R0 >>= 9;               /* R0 = current MSEL */

    /* get current VCO */
    IMM32(R6,CLKIN_Hz);     /* R6 = CLKIN [Hz] */
    R0 *= R6;               /* R0 = MSEL * CLKIN */
    R6 = DF (z);            /* R6 = DF bit */
    R6 = R6 & R7;           /* R6 = DF bit [1/0] */
    R0 >>= R6;              /* R0 = current VCO [Hz] */


    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__get_vco_hz.end:


/****************************************************************************
 Name:                  __get_cclk_hz
 Description:           get core clock frequency in Hz
 Input Parameters:      none
 Return Parameters:     R0 = CCLK
 Registers Used:        R7:6,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__get_cclk_hz:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    R7 = w[P5 + lo(PLL_CTL)] (z);
    CC = bittst(R7,bitpos(BYPASS));
    IMM32(R0,CLKIN_Hz);
    if CC jump get_cclk_hz.skip;

    call __get_vco_hz;       /* R0 = current VCO [Hz] */

    /* get current CSEL */
    R7 = w[P5 + lo(PLL_DIV)] (z);
    R6 = CSEL (z);
    R7 = R7 & R6;
    R7 >>= 4;               /* R7 = current CSEL */

    R0 >>= R7;              /* R0 = current CCLK */


    get_cclk_hz.skip:
    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__get_cclk_hz.end:


/****************************************************************************
 Name:                  __get_sclk_hz
 Description:           get system clock frequency in Hz
 Input Parameters:      none
 Return Parameters:     R0 = SCLK
 Registers Used:        R1,R7,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__get_sclk_hz:

    link 0;
    [--SP] = ASTAT;
    [--SP] = R7;
    [--SP] = R1;
    [--SP] = P5;
    IMM32(P5,SYS_MMR_BASE);


    R7 = w[P5 + lo(PLL_CTL)] (z);
    CC = bittst(R7,bitpos(BYPASS));
    IMM32(R0,CLKIN_Hz);
    if CC jump get_sclk_hz.skip;

    call __get_vco_hz;       /* R0 = current VCO [Hz] */

    /* get current SSEL */
    R1 = w[P5 + lo(PLL_DIV)] (z);
    R7 = SSEL (z);
    R1 = R1 & R7;           /* R1 = current SSEL */

    /* get current SCLK */
    udiv32(); /* Registers passed: Dividend = R0, Divisor = R1, Quotient -> R0 */
                            /* R0 = current SCLK */


    get_sclk_hz.skip:
    P5 = [SP++];
    R1 = [SP++];
    R7 = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__get_sclk_hz.end:


#endif /* defined __USEBFSYSCONTROL__ */


SDC:


#if defined (__ADSPBF54x__) && defined (SDRAM_SIZE)


/****************************************************************************
 Name:                  __sdram_en
 Description:           enable DDR-SDRAM
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__sdram_en:
 
    link 0;
    [--SP] = ASTAT;
    [--SP] = R7;
    [--SP] = P5;
    IMM32(P5,SYS_MMR_BASE);


    /****************************************************************************
     Programming Model:
     Access to the DDR controller registers ONLY can be made after releasing the
     DDR controller soft reset bit in the reset control register by writing a 1
     in bit[0] in the register.
     The user may write to the DDR control registers as long as the controller
     is not accessing memory devices. Otherwise, the controller responds to any
     writes to its registers after it finishes any ongoing memory accesses.
    ****************************************************************************/


    /*************************
    * Reset Control Register *
    **************************/

    R7 = w[P5 + lo(EBIU_RSTCTL)];
    bitset(R7,bitpos(DDRSRESET));
    w[P5 + lo(EBIU_RSTCTL)] = R7;


    /****************************
    * Memory Control 0 Register *
    ****************************/

    IMM32(R7,EBIU_DDRCTL0_VAL);
    [P5 + lo(EBIU_DDRCTL0)] = R7;


    /****************************
    * Memory Control 1 Register *
    ****************************/

    IMM32(R7,EBIU_DDRCTL1_VAL);
    [P5 + lo(EBIU_DDRCTL1)] = R7;


    /****************************
    * Memory Control 2 Register *
    ****************************/

    IMM32(R7,EBIU_DDRCTL2_VAL);
    [P5 + lo(EBIU_DDRCTL2)] = R7;


    P5 = [SP++];
    R7 = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__sdram_en.end:


#endif /* defined (__ADSPBF54x__) && defined (SDRAM_SIZE) */


#if !defined (__ADSPBF54x__) && defined (SDRAM_SIZE)


/****************************************************************************
 Name:                  __sdram_en
 Description:           enable SDR-SDRAM
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__sdram_en:
 
    link 0;
    [--SP] = ASTAT;
    [--SP] = R7;
    [--SP] = P5;
    IMM32(P5,SYS_MMR_BASE);


    /*****************************************************************************
     Description:   Configure and enable SDRAM

     1. Ensure the clock to the SDRAM is stable after the power has stabilized
        for the proper amount of time (typically 100 us).
     2. Write to the SDRAM refresh rate control register (EBIU_SDRRC).
     3. Write to the SDRAM memory bank control register (EBIU_SDBCTL).
     4. Write to and SDRAM memory global control register (EBIU_SDGCTL).
     5. Perform SDRAM access.
    ******************************************************************************/


    /********************************
    * SDRAM Control Status Register *
    ********************************/

    PollSdcIdle:
    R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
    CC = bittst(R7,bitpos(SDCI));
    if !CC jump PollSdcIdle;

    /* already enabled? */
    R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
    CC = bittst(R7, bitpos(SDRS));
    if !CC jump SkipSdcPowerUp;

    /* clear SDRAM EAB sticky error status (W1C) */
    R7 = SDEASE(z);
    w[P5 + lo(EBIU_SDSTAT)] = R7;


    /***************************************************************************
    * SDRAM Refresh Rate Control Register                                      *
    * RDIV has always to be re-calculated according to the actual system clock *
    * You can either use the defined value that fits to the PCC settings or    *
    * call __get_rdiv to calculate RDIV 'on the fly'                            *
    ***************************************************************************/

    #if defined EBIU_SDRRC_VAL
        R0.L = EBIU_SDRRC_VAL;
    #else
        call __get_rdiv;
    #endif
    w[P5 + lo(EBIU_SDRRC)] = R0;


    /*************************************
    * SDRAM Memory Bank Control Register *
    *************************************/

    #if defined __ADSPBF561__
        IMM32(R7,EBIU_SDBCTL_VAL);
        [P5 + lo(EBIU_SDBCTL)] = R7;
    #else
        R7.L = EBIU_SDBCTL_VAL;
        w[P5 + lo(EBIU_SDBCTL)] = R7;
    #endif


    /***************************************
    * SDRAM Memory Global Control Register *
    ***************************************/

    IMM32(R7,EBIU_SDGCTL_VAL);
    [P5 + lo(EBIU_SDGCTL)] = R7;


    /******************************
    * Finalize SDC initialization *
    ******************************/

    /* a transfer is required to finalize SDC initialization ! */
    IMM32(P4,0x4);
    nop;
    R7 = w[P4] (z);

    PollSdcPowerUpFinished:
    R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
    CC = bittst(R7,bitpos(SDRS));
    if CC jump PollSdcPowerUpFinished;


    SkipSdcPowerUp: nop;
    P5 = [SP++];
    R7 = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__sdram_en.end:


#endif /* !defined (__ADSPBF54x__) && defined (SDRAM_SIZE) */


#if !defined (__ADSPBF54x__) && defined (SDRAM_SIZE)


/****************************************************************************
 Name:                  __get_rdiv
 Description:           Calculate RDIV value for SDRAM Refresh Rate Control Register
                        RDIV = ( ( SCLK[MHz] * tREF[ms] ) / NRA ) - ( tRAS + tRP ) [clock cycles]
                        RDIV = ( ( SCLK * 10^6 * 1/s * tREF * 10^-3 s ) / NRA ) - ( tRAS + tRP ) [clock cycles]
                        SCLK = ( CLKIN * MSEL ) / SSEL = VCO / SSEL
                        VCO = get_vco_hz
                        SCLK = get_sclk_hz
                        MSEL = Extracted from PLL_CTL register
                        SSEL = Extracted from PLL_DIV register
                        tREF = Definition
                        NRA  = Definition
                        tRAS = Extracted from EBIU_SDGCTL
                        tRP = Extracted from EBIU_SDGCTL
 Input Parameters:      none
 Return Parameters:     R0 = RDIV_VAL
 Registers Used:        R7:1
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__get_rdiv:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:1);


    call __get_sclk_hz;      /* R0 = current SCLK [Hz] */

    R1 = NRA (z);
    R7 = 1000 (z);
    R1 *= R7;
    udiv32();

    R7 = tREF (z);          /* R7 = tREF [ms] */
    R0 *= R7;               /* R0 = ( SCLK / NRA ) * tREF */

    IMM32(R7,EBIU_SDGCTL_VAL);

    /* get tRP */
    R5 = TRP;
    R5 = R5 & R7;
    R5 >>= 11;              /* R5 = tRP */

    /* get tRAS */
    R6 = TRAS;
    R6 = R6 & R7;
    R6 >>= 6;               /* R6 = tRAS */

    R7 = R5 + R6;           /* R7 = tRAS + tRP */

    R0 = R0 - R7;           /* R0 = RDIV */


    (R7:1) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__get_rdiv.end:


#endif /* !defined (__ADSPBF54x__) && defined (SDRAM_SIZE) */


#if defined SDRAM_SIZE


SDC_SELFREFRESH:


/****************************************************************************
 Name:                  __sdram.self_refresh.enter
 Description:           enter SDC self-refresh
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5:4,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__sdram.self_refresh.enter:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:4);
    IMM32(P5,SYS_MMR_BASE);


    /* SDRAM enabled? */
    #if defined __ADSPBF54x__
        R7 = w[P5 + lo(EBIU_RSTCTL)] (z);
        CC = bittst(R7,bitpos(DDRSRESET));
    #else
        R7 = w[P5 + lo(EBIU_SDBCTL)] (z);
        CC = bittst(R7,bitpos(EBE));
    #endif
    if !CC jump sdram.self_refresh.enter.done;


    /* SDRAM in self-refresh? */
    #if defined __ADSPBF54x__
        R7 = w[P5 + lo(EBIU_RSTCTL)] (z);
        CC = bittst(R7,bitpos(SRACK));
    #else
        R7 = w[P5 + lo(EBIU_SDBCTL)] (z);
        CC = bittst(R7,bitpos(SDSRA));
    #endif
    if CC jump sdram.self_refresh.enter.done;


    /* a transfer is required to finalize SDC initialization ! */
    /* Already done in: init_code examples and 'bfstartup.asm' */
    IMM32(P4,SDRAM_BANK0_START);
    nop;
    R7 = w[P4] (z);
    ssync; /* force last SDRAM access to finish */


    #if defined __ADSPBF54x__

        /* enter self-refresh mode */
        R7 = w[P5 + lo(EBIU_RSTCTL)];
        bitset(R7,bitpos(SRREQ));
        w[P5 + lo(EBIU_RSTCTL)] = R7;
        ssync;

        /* poll self-refresh status */
        enter.PollSelfRefreshStatus:
        R7 = w[P5 + lo(EBIU_RSTCTL)];
        ssync;
        CC = bittst(R7,bitpos(SRACK));
        if !CC jump enter.PollSelfRefreshStatus;

    #else

        enter.PollSdcIdle:
        R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
        CC = bittst(R7,bitpos(SDCI));
        if !CC jump enter.PollSdcIdle;

        /* enter self-refresh mode */
        R7 = [P5 + lo(EBIU_SDGCTL)];
        bitset(R7,bitpos(SRFS));
        [P5 + lo(EBIU_SDGCTL)] = R7;
        ssync;

        /* poll self-refresh status */
        enter.PollSelfRefreshStatus:
        R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
        ssync;
        CC = bittst(R7,bitpos(SDSRA));
        if !CC jump enter.PollSelfRefreshStatus;

    #endif


    sdram.self_refresh.enter.done:
    (R7:7,P5:4) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__sdram.self_refresh.enter.end:


/****************************************************************************
 Name:                  __sdram.self_refresh.leave
 Description:           leave SDC self-refresh
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5:4,ASTAT
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__sdram.self_refresh.leave:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:4);
    IMM32(P5,SYS_MMR_BASE);


    /* SDRAM enabled? */
    #if defined __ADSPBF54x__
        R7 = w[P5 + lo(EBIU_RSTCTL)];
        CC = bittst(R7,bitpos(DDRSRESET));
    #else
        R7 = w[P5 + lo(EBIU_SDBCTL)] (z);
        CC = bittst(R7,bitpos(EBE));
    #endif
    if !CC jump sdram.self_refresh.leave.done;


    /* SDRAM in self-refresh? */
    #if defined __ADSPBF54x__
        R7 = w[P5 + lo(EBIU_RSTCTL)];
        CC = bittst(R7,bitpos(SRACK));
    #else
        R7 = w[P5 + lo(EBIU_SDBCTL)] (z);
        CC = bittst(R7,bitpos(SDSRA));
    #endif
    if !CC jump sdram.self_refresh.leave.done;


    #if defined __ADSPBF54x__

        R7 = w[P5 + lo(EBIU_RSTCTL)] (z);
        bitclr(R7,bitpos(SRREQ));
        w[P5 + lo(EBIU_RSTCTL)] = R7;
        ssync;

        leave.PollSelfRefreshStatus:
        R7 = w[P5 + lo(EBIU_RSTCTL)] (z);
        ssync;
        CC = bittst(R7,bitpos(SRACK));
        if CC jump leave.PollSelfRefreshStatus;

    #else

        leave.PollSdcIdle:
        R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
        CC = bittst(R7,bitpos(SDCI));
        if !CC jump leave.PollSdcIdle;

        leave.PollSdcSR:
        R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
        CC = bittst(R7,bitpos(SDSRA));
        if !CC jump leave.PollSdcSR;

        R7 = [P5 + lo(EBIU_SDGCTL)];
        bitclr(R7,bitpos(SRFS));
        [P5 + lo(EBIU_SDGCTL)] = R7;
        ssync;

         /* a transfer is required to finalize SDC initialization ! */
        IMM32(P4,SDRAM_BANK0_START);
        nop;
        R7 = w[P4] (z);
        #if defined DATA_CACHE_EN
            #if defined LIMIT32kB
            IMM32(P4,0x2000);
            R7 = w[P4];
            IMM32(P4,0x4000);
            R7 = w[P4];
            IMM32(P4,0x6000);
            R7 = w[P4];
            #else
            IMM32(P4,SDRAM_BANK1_START);
            R7 = w[P4];
            IMM32(P4,SDRAM_BANK2_START);
            R7 = w[P4];
            IMM32(P4,SDRAM_BANK3_START);
            R7 = w[P4];
            #endif
        #endif

        leave.PollSelfRefreshStatus:
        R7 = w[P5 + lo(EBIU_SDSTAT)] (z);
        ssync;
        CC = bittst(R7,bitpos(SDSRA));
        if CC jump leave.PollSelfRefreshStatus;

    #endif


    sdram.self_refresh.leave.done:
    (R7:7,P5:4) = [SP++];
    ASTAT = [SP++];
    unlink;
    rts;

__sdram.self_refresh.leave.end:


#endif /* SDRAM_SIZE */


RESETS:


/****************************************************************************
 Name:                  __sysreset
 Description:           issue system software reset
                        SDC -> Self-Refresh Mode
 Result:                Resets only the peripherals, excluding the RTC
                        (Real-Time Clock) block and most of the DPMC.
                        The system software reset clears bits[15:13]
                        and bits[11:4] of the SYSCR register,
                        but not the WURESET bit.
                        Does not reset the core.
                        Does not initiate a boot sequence.
                        Sequencing continues at the instruction after the SWRST write.
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__sysreset:

    #if defined __USEBFSYSCONTROL__
        link sizeof(ADI_SYSCTRL_VALUES)+2;
    #else
        link 0;
    #endif
    [--SP] = (R7:0,P5:0);
    #if defined __USEBFSYSCONTROL__
        SP += -12;
    #endif

    IMM32(P5,SYS_MMR_BASE);


    #if defined SDRAM_SIZE
//        call __sdram.self_refresh.enter;
    #endif


    #if defined (__USEBFSYSCONTROL__) && !defined (__WORKAROUND_05000353__)
        R0 = ( SYSCTRL_SYSRESET );
        R1 = 0 (z);
        R2 = 0 (z);
        IMM32(P4,BFROM_SYSCONTROL);
        call(P4);
    #else
        /* Issue System Software Reset */
        R7 = SYSTEM_RESET (z);
        #if defined __ADSPBF561__
            w[P5 + lo(SICA_SWRST)] = R7;
        #else
            w[P5 + lo(SWRST)] = R7;
        #endif
        ssync;

        /**********************************************************************
         Wait for system reset to complete (needs to be 5 SCLKs). Assuming
         a worst case CCLK:SCLK ratio (15:1), use 5*15 = 75 as the loop count.
        **********************************************************************/
        P4 = 75;
        lsetup(sysreset.lp,sysreset.lp) LC0 = P4;
        sysreset.lp: nop;

        /* Clear System Software Reset */
        R7 = 0x0 (z);
        #if defined __ADSPBF561__
            w[P5 + lo(SICA_SWRST)] = R7;
        #else
            w[P5 + lo(SWRST)] = R7;
        #endif
        ssync;
    #endif


    /* Clear System Software Reset Status */
    /* Workaround for the Workaround !!! */
    #if defined __WORKAROUND_05000353__
        R7 = w[P5 + lo(SWRST)];
        ssync;
    #endif


    #if defined SDRAM_SIZE
        call __sdram_en;

        IMM32(P4,SDRAM_BANK0_START);
        nop;
        R7 = w[P4] (z);
    #endif


    #if defined __USEBFSYSCONTROL__
        SP += 12;
    #endif
    (R7:0,P5:0) = [SP++];
    unlink;
    rts;

__sysreset.end:


/****************************************************************************
 Function:              __softreset
 Description:           issue system software reset + core-only software reset
                        SDC -> Self-Refresh Mode
 Result:                Does a system software reset
                        Does reset the core.
                        Does initiate a boot sequence.
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__softreset:

    #if defined __USEBFSYSCONTROL__
        link sizeof(ADI_SYSCTRL_VALUES)+2;
    #else
        link 0;
    #endif
    [--SP] = (R7:0,P5:0);
    #if defined __USEBFSYSCONTROL__
        SP += -12;
    #endif


    #if defined (__USEBFSYSCONTROL__) && !defined (__WORKAROUND_05000353__)
        #if defined SDRAM_SIZE
//            call __sdram.self_refresh.enter;
        #endif
        R0 = ( SYSCTRL_SOFTRESET );
        R1 = 0 (z);
        R2 = 0 (z);
        IMM32(P4,BFROM_SYSCONTROL);
        call(P4);
    #else
        /* Issue System Software Reset */
        call __sysreset;

        /* Issue Core-Only Software Reset - forces reboot */
        raise 1;
    #endif


    #if defined __USEBFSYSCONTROL__
        SP += 12;
    #endif
    (R7:0,P5:0) = [SP++];
    unlink;
    rts;

__softreset.end:


MISC:


/****************************************************************************
 Name:                  __trace_unit.enable
 Description:           enable Trace Unit
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7, P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__trace_unit.enable:

    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,CORE_MMR_BASE);

    R7 = [P5 + lo(TBUFCTL)];
//    bitset(R7,bitpos(TBUFPWR));
//    bitset(R7,bitpos(TBUFEN));
    bitset(R7,0);
    bitset(R7,1);
    [P5 + lo(TBUFCTL)] = R7;

    (R7:7,P5:5) = [SP++];
    unlink;
    rts;

__trace_unit.enable.end:


/****************************************************************************
 Name:                  __trace_unit.disable
 Description:           disable Trace Unit
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7, P5
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__trace_unit.disable:

    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,CORE_MMR_BASE);

    R7 = [P5 + lo(TBUFCTL)];
//    bitclr(R7,bitpos(TBUFPWR));
//    bitclr(R7,bitpos(TBUFEN));
    bitclr(R7,0);
    bitclr(R7,1);
    [P5 + lo(TBUFCTL)] = R7;

    (R7:7,P5:5) = [SP++];
    unlink;
    rts;

__trace_unit.disable.end:


/****************************************************************************
 Name:                  __profiling.enable
 Description:           reset and enable cycle counter
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__profiling.enable:

    link 0;
    [--SP] = R7;

    R7 = 0;
    CYCLES = R7;
    CYCLES2 = R7;
    R7 = SYSCFG;
    nop;
    bitset(R7,1);
//    bitset(R7,bitpos(CCEN));
    SYSCFG = R7;

    R7 = [SP++];
    unlink;
    rts;

__profiling.enable.end:


/****************************************************************************
 Name:                  __profiling.disable
 Description:           disable cycle counter
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7
 Global Registers Used: none
 C-Callable :           no
*****************************************************************************/

__profiling.disable:

    link 0;
    [--SP] = R7;

    R7 = SYSCFG;
    nop;
    bitclr(R7,1);
//    bitclr(R7,bitpos(CCEN));
    SYSCFG = R7;

    R7 = [SP++];
    unlink;
    rts;

__profiling.disable.end:


/*****************************************************************************
 EOF
******************************************************************************/
