/********************************************************/
/*	Exception Handler for DCPLB Miss		 			*/
/*														*/
/*	The Exception Handler does the following:			*/
/*  (a) Veriy that it is a DCPLB Miss Exception. 		*/
/*	(b) Extract the faulting address. 					*/
/*	(c) Search for the entry containing above address	*/
/*	    in the DCPLB table.								*/
/*	(d) Replace entry in one of the DPLB descriptors 	*/ 
/*	    with the new entry.								*/
/*	(e) The replacement policy uses:					*/
/*		(i) Round Robin Scheduling						*/
/*		(ii) Checks for the CPLB_LOCK					*/
/*														*/
/*							Initial Version				*/
/*							Kunal Singh					*/
/*							DSP applications, ADI		*/
/*							Dated 3 April, 2005			*/
/*														*/
/********************************************************/

#include <blackfin.h>
#include <services_types.h>
#include <stdio.h>
#include<sys/exception.h>
#include "main.h"


volatile unsigned long Exception_source;
volatile unsigned long cache;
volatile unsigned long Fault_Addr;
volatile unsigned long Search_CPLB_entry;
volatile unsigned long *address_pointer, *data_pointer;
volatile unsigned long *temp_pointer;
volatile unsigned long table_index;
volatile unsigned long Start_address, End_address;
volatile unsigned long replace_index;
volatile unsigned long Validate_CPLB_replace;
unsigned int tmp;
/*
#undef pDCPLB_ADDR0
#define pDCPLB_ADDR0 ((volatile unsigned long *)DCPLB_ADDR0)
#undef pDCPLB_FAULT_ADDR
#define pDCPLB_FAULT_ADDR ((volatile unsigned long *)DCPLB_FAULT_ADDR)
*/

/************************************************/
/*      Exception Handler                       */
/************************************************/
EX_EXCEPTION_HANDLER(exception_isr)
{

    /****************************/
    /* Read the Exception Cause */
    /****************************/
    asm("r0 = SEQSTAT;");
    asm("P1.h = _Exception_source;");
    asm("P1.l = _Exception_source;");
    asm("W[P1] = r0;");


    if((Exception_source&0x3F) == 0x26)
    /************************/
    /* Verify a DCPLB Miss  */
    /************************/
    {
        /* Read the faulting Address */
        Fault_Addr = (unsigned long) *pDCPLB_FAULT_ADDR;

        Search_CPLB_entry = 1;
        table_index = 0;

        address_pointer = D_cplb_addrs;
        data_pointer = D_cplb_data;

        /************************************/
        /* Search for the Faulting Address  */
        /* in the DCPLB Page Table          */
        /************************************/
        while(Search_CPLB_entry)
        {
            Start_address = *(address_pointer + table_index);

            tmp = ((*(data_pointer + table_index))&(0x00030000)); 
            tmp = (tmp>>16);

            if (tmp == 0) { End_address = Start_address + 1024; }
            if (tmp == 1) { End_address = Start_address + 4096; }
            if (tmp == 2) { End_address = Start_address + 1048576; }
            if (tmp == 3) { End_address = Start_address + 4194304; }

            if ((Fault_Addr >= Start_address)&(Fault_Addr < End_address)) { Search_CPLB_entry = 0; }
            table_index ++;
        }

        /****************************************************/
        /* This is the index of entry in DCPLB Page Table   */
        /* which correspond to the faulting address         */
        /****************************************************/
        table_index = table_index - 1;


        /****************************************************/
        /* replace_index is the index of DCPLB descriptor   */
        /* to be replaced by the new Entry                  */
        /*                                                  */
        /* This is based on Round Robin Scheduling          */
        /****************************************************/
        Validate_CPLB_replace = 1;
        temp_pointer = (pDCPLB_DATA0) + replace_index;

        /********************************************************/
        /* Check CPLB_LOCK bit of the descriptor to be replaced */
        /* Change the replace_index if the CPLB is locked       */
        /********************************************************/
        while(Validate_CPLB_replace)
        {
            tmp = *temp_pointer;
            tmp = (tmp) & (0x2);
            tmp = (tmp>> 1);
            if (tmp == 1)
            {
                /* The CPLB is locked */
                replace_index += 1;
                temp_pointer +=1;
            }
            else
            {
                Validate_CPLB_replace = 0;
            }
        }

        /*************************************/
        /* Replace the CPLB at replace_index */
        /*************************************/
        /* Disable the DCPLBs */
        *pDMEM_CONTROL &= 0xFFFFFFFD;
        asm("ssync;");

        /* Replace the CPLB */
        *(pDCPLB_DATA0 + replace_index) = *(data_pointer + table_index);
        *(pDCPLB_ADDR0 + replace_index) = (volatile void **) *(address_pointer + table_index);
        asm("ssync;");

        /* Enable the DCPLBs */
        *pDMEM_CONTROL |= 0x2;
        asm("ssync;");

        /********************************/
        /* Increment the replace index  */
        /* Round Robin Scheduling       */
        /********************************/
        replace_index += 1;
    }
}
